/*++
Copyright (c) 1994-2000 Sangoma Technologies Inc.
Module Name:
    sdla.c
Abstract:
    S502/S503/S508/S514 SDLA kernel-mode driver for Windows NT Workstation 3.5 and later
    May support multiple devices (sdla0, sdla1, ... sdla15)
Environment:
    kernel mode only
Notes:
	bypasses resource assignment
Revision History:

v2.0.0.1	26/08/02: Added autodetection option of one S5141 card	-David-
			or one S5142 card. If S5141 is foind one SDLA card is
			configured, if S5142 is found then 2 SDLA cards configured
			automatically.

v2.0.0.0	14/08/01: Added autodetection option of one S514 card	-David-
			Changed (SDLA_BLOCKOP *)->offset from USHORT to ULONG
			in "sdla.h" to use with new Frame Relay code on S514 
			card for Educational kit. All ?api32mc DLLs must be 
			recopiled with new "sdla.h" to avoid kernel	mode traps.

v1.0.0.4	12/01/00: Added logging of errors to Sdlalog.wri file	-David-
			03/31/00: Added S514 PCI support						-David-

v1.0.0.0	08/07/96: Added resource assignment
			Trying to lock down user-mode paged memory pointed to by
			(SDLA_BLOCKOP *)->buffer I think it may be causing page
			faults without doing this.
			I aborted this.

--*/


#include "ntddk.h"
#include "string.h"
#include "sdla.h"
#include "sdlapci.h"
#include <stdio.h>//for sprinf()
#include <wanpipe\sdla_front_end.h>

// define this to get some debug messages
//#define DEBG

#define OFF_Z80_WIN_SELECT 		0x02	// the offset from the card I/O port base address of the Z80 window

// used for storing log strings
char globalLogString[300];

// 16-bit ported stuff
PUCHAR cardmem[MAX_CARDS];
PUCHAR portmem[MAX_CARDS];
USHORT ports[MAX_CARDS];

PUCHAR stop_start_s514_register[MAX_CARDS];

ULONG reserve_mem[MAX_CARDS];
ULONG reserve_port[MAX_CARDS];
USHORT reserve_window_size[MAX_CARDS];

ULONG reserve_bus[MAX_CARDS];
ULONG reserve_slot[MAX_CARDS];
ULONG reserve_CPU[MAX_CARDS];
BOOLEAN dont_use[MAX_CARDS];

USHORT hw_type[MAX_CARDS];

/* bit record of available cards */
USHORT available_cards = 0;

WCHAR *fool_proof[] = { L"sdla0", L"sdla1", L"sdla2", L"sdla3",
			L"sdla4", L"sdla5", L"sdla6", L"sdla7",
			L"sdla8", L"sdla9", L"sdla10", L"sdla11",
			L"sdla12", L"sdla13", L"sdla14", L"sdla15" };


/* number of times any given application registers for use of the
particular card...
if it is zero on registration, a selector and pointer to the card's
base memory is set up (if succeeds); if it falls to zero by
deregistration, the selector is freed
*/
USHORT cardinstances[MAX_CARDS];
UCHAR blockregistration[MAX_CARDS];

ULONG temp_io_port, temp_window_segment, temp_window_offset, temp_window_size;

RTL_QUERY_REGISTRY_TABLE rtable[] = {
{ NULL, RTL_QUERY_REGISTRY_DIRECT | RTL_QUERY_REGISTRY_REQUIRED, L"io_port", (PVOID)&temp_io_port, 0, NULL, 0 },
{ NULL, RTL_QUERY_REGISTRY_DIRECT | RTL_QUERY_REGISTRY_REQUIRED, L"window_segment", (PVOID)&temp_window_segment, 0, NULL, 0 },
{ NULL, RTL_QUERY_REGISTRY_DIRECT | RTL_QUERY_REGISTRY_REQUIRED, L"window_offset", (PVOID)&temp_window_offset, 0, NULL, 0 },
{ NULL, RTL_QUERY_REGISTRY_DIRECT | RTL_QUERY_REGISTRY_REQUIRED, L"window_size", (PVOID)&temp_window_size, 0, NULL, 0 },
{ NULL, 0, NULL, NULL, 0, NULL, 0 }
};

ULONG cardmem_bytes[MAX_CARDS];

USHORT BusNumber, SlotNumber, CPU;

RTL_QUERY_REGISTRY_TABLE rtable514[] = {
{ NULL, RTL_QUERY_REGISTRY_DIRECT | RTL_QUERY_REGISTRY_REQUIRED, L"BusNumber", (PVOID)&BusNumber, 0, NULL, 0 },
{ NULL, RTL_QUERY_REGISTRY_DIRECT | RTL_QUERY_REGISTRY_REQUIRED, L"SlotNumber", (PVOID)&SlotNumber, 0, NULL, 0 },
{ NULL, RTL_QUERY_REGISTRY_DIRECT | RTL_QUERY_REGISTRY_REQUIRED, L"CPU", (PVOID)&CPU, 0, NULL, 0 },
{ NULL, 0, NULL, NULL, 0, NULL, 0 }
};

//
//If set to 1, only one S514 PCI card will be used (and nothing else),
//the card must be detected by the driver without help from the "detection" driver.
//
USHORT AutoConfigS514 = 0;
BOOLEAN adapterFound = FALSE;

RTL_QUERY_REGISTRY_TABLE auto_config_S514_table[] = {
{ NULL, RTL_QUERY_REGISTRY_DIRECT | RTL_QUERY_REGISTRY_REQUIRED, L"AutoConfigS514", (PVOID)&AutoConfigS514, 0, NULL, 0 },
{ NULL, 0, NULL, NULL, 0, NULL, 0 }
};

/////////////////////////////////////////////////////////////////////////////
typedef struct _KDVR_DEVICE_EXTENSION {
    ULONG Information;
} KDVR_DEVICE_EXTENSION, *PKDVR_DEVICE_EXTENSION;

PDRIVER_OBJECT ThisDriverObject;

//
// Function declarations
//
static NTSTATUS KdvrDispatch(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp);
static NTSTATUS KdvrDispatchIoctl(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp);
VOID KdvrUnloadDriver(IN PDRIVER_OBJECT DriverObject);
UCHAR MapMemory(UCHAR cardnumber);
void UnmapMemory(UCHAR cardnumber);
void FormRegistryPath(PUNICODE_STRING p, PWCH buffer, UCHAR cardnumber);
BOOLEAN ReserveResources(IN PDRIVER_OBJECT DriverObject);
void FreeResources(IN PDRIVER_OBJECT DriverObject);

static VOID SANGPrintResourceList(PCM_RESOURCE_LIST Resources);
static VOID SANGPrintConfig(PPCI_COMMON_CONFIG  configInfo);
BOOLEAN InitializeCard514(UCHAR cardnumber);
ULONG MapSetZ80Window(ULONG logical_offset, UCHAR cardnumber);
void StartUpLog(void);
BOOLEAN DoMemoryTest(UCHAR cardnumber);
VOID FindSangomaCard(void);
VOID SangReturnPool(PPCI_COMMON_CONFIG  ConfigInfo);

// static UNICODE_STRING rpathString, classString;

NTSTATUS DriverEntry(IN PDRIVER_OBJECT DriverObject,
			IN PUNICODE_STRING RegistryPath)
{
	UNICODE_STRING nameString, linkString;
	PDEVICE_OBJECT deviceObject;
	NTSTATUS status;
	PKDVR_DEVICE_EXTENSION extension;
	ULONG bytes, iobytes;
	UCHAR i;

	PVOID pv;

	WCHAR tw[80];
	UNICODE_STRING path1;

	RtlInitUnicodeString(&nameString, L"\\Device\\sdla");

	status = IoCreateDevice( DriverObject,
			     sizeof(KDVR_DEVICE_EXTENSION),
			     &nameString,
			     FILE_DEVICE_UNKNOWN,
			     0,
			     FALSE,
			     &deviceObject );

	if (!NT_SUCCESS(status))
	{	sprintf(globalLogString, "IoCreateDevice() Failed.\0");
		StartUpLog();
		return(status);
	}

	deviceObject->Flags |= DO_BUFFERED_IO;

    RtlInitUnicodeString(&linkString, L"\\DosDevices\\SDLA");
    status = IoCreateSymbolicLink(&linkString, &nameString);

    if (!NT_SUCCESS(status))
	{	sprintf(globalLogString, "IoCreateSymbolicLink() Failed.\0");
		StartUpLog();

		IoDeleteDevice(DriverObject->DeviceObject);
		return(status);
    }

    DriverObject->DriverUnload = KdvrUnloadDriver;
    DriverObject->MajorFunction[IRP_MJ_CREATE] = KdvrDispatch;
    DriverObject->MajorFunction[IRP_MJ_CLOSE]  = KdvrDispatch;
    DriverObject->MajorFunction[IRP_MJ_INTERNAL_DEVICE_CONTROL] = KdvrDispatchIoctl;
    DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL] = KdvrDispatchIoctl;
    extension = deviceObject->DeviceExtension;

    extension->Information = 0x12345678;

	for (i=0; i < MAX_CARDS; i++)
		hw_type[i] = 508;

	//Read autodetection flag
	{
		tw[0] = 0;
		RtlInitUnicodeString(&path1, tw);
		path1.MaximumLength = 79;
		path1.Length = 0;
		
		RtlAppendUnicodeToString(&path1, L"SDLA\\Parameters\\");
		RtlQueryRegistryValues(RTL_REGISTRY_SERVICES, path1.Buffer, &auto_config_S514_table[0], NULL, NULL);

#ifdef DEBG
	    DbgPrint("AutoConfigS514 == %d\n", AutoConfigS514);
#endif	
	}

	//
	//find S5141 0r S5142 card if autodetection flag is set
	//
	if(AutoConfigS514 == 1)
		FindSangomaCard();
	
	if(AutoConfigS514 == 1 && adapterFound == FALSE)
	{	//No S514 card detected. Log error and exit.
		sprintf(globalLogString, 
			"Configured for autodetection of S514 PCI card but no S5141 or S5142 card was found. Please make sure S514 is properly inserted into the slot.\0");
		StartUpLog();

		IoDeleteDevice (DriverObject->DeviceObject);
		IoDeleteSymbolicLink (&linkString);
		return(STATUS_INSUFFICIENT_RESOURCES);

	}else if(AutoConfigS514 == 0){	//it is 0
		/* no one registered a card yet */
		for (i=0; i < MAX_CARDS; i++)
		{
			cardinstances[i] = 0;
			cardmem[i] = NULL;
			blockregistration[i] = 0;
			stop_start_s514_register[i] = NULL;
			dont_use[i] = FALSE;
			
			FormRegistryPath(&path1, tw, i);
			
			if (RtlQueryRegistryValues(RTL_REGISTRY_SERVICES, path1.Buffer, &rtable514[0], NULL, NULL)
				== STATUS_SUCCESS)
			{
				available_cards |= (0x0001 << i);
				reserve_bus[i] = BusNumber;
				reserve_slot[i] = SlotNumber;
				reserve_CPU[i] = CPU;
				hw_type[i] = 514;
			}
			
			if(hw_type[i] != 514)
			{
				if (RtlQueryRegistryValues(RTL_REGISTRY_SERVICES, path1.Buffer, &rtable[0], NULL, NULL)
					== STATUS_SUCCESS)
				{
					available_cards |= (0x0001 << i);
					reserve_port[i] = temp_io_port;
					reserve_mem[i] = (temp_window_segment << 4) + temp_window_offset;
					reserve_window_size[i] = (USHORT)temp_window_size;
				}
			}
		}
	}

	ThisDriverObject = DriverObject;
	
	if (ReserveResources(DriverObject) != TRUE)
	{
		IoDeleteDevice (DriverObject->DeviceObject);
		IoDeleteSymbolicLink (&linkString);
		return(STATUS_INSUFFICIENT_RESOURCES);
	}
#ifdef DEBG
    DbgPrint("Successfully finished SDLA DriverEntry routine ...\n");
#endif	
    return(STATUS_SUCCESS);
}

VOID KdvrUnloadDriver(IN PDRIVER_OBJECT DriverObject)
{
	UNICODE_STRING nameString, linkString;

	FreeResources(DriverObject);

	IoDeleteDevice (DriverObject->DeviceObject);

	RtlInitUnicodeString(&linkString, L"\\DosDevices\\SDLA");
	RtlInitUnicodeString(&nameString, L"\\Device\\sdla");

	IoDeleteSymbolicLink (&linkString);

}

static NTSTATUS KdvrDispatch(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp)
{
    NTSTATUS status;
    PIO_STACK_LOCATION irpSp;

    UNREFERENCED_PARAMETER(DeviceObject);
    irpSp = IoGetCurrentIrpStackLocation(Irp);

    switch (irpSp->MajorFunction)
	{
	case IRP_MJ_CREATE:
	Irp->IoStatus.Status = STATUS_SUCCESS;
	    Irp->IoStatus.Information = 0L;
	    break;

	case IRP_MJ_CLOSE:
	    Irp->IoStatus.Status = STATUS_SUCCESS;
	    Irp->IoStatus.Information = 0L;
	    break;
    }

    status = Irp->IoStatus.Status;
    IoCompleteRequest(Irp, 0);
    return(status);
}


/*
NOTE!!!
all card-related calls (register,prepare,deregister,complete,getcards)
must be exclusive as they all share data (read and write)
all other calls share cardmem/portmem for reading only
*/
static NTSTATUS KdvrDispatchIoctl(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp)
{
    NTSTATUS status;
    PIO_STACK_LOCATION irpSp;
    PKDVR_DEVICE_EXTENSION extension = DeviceObject->DeviceExtension;
	UCHAR x;
	XFER_BYTE_CONTROL *OutputBuffer;
	SDLA_BLOCKOP *block;
	CARD_TRANSACTION *trans;
	UCHAR cardnumber;
	ULONG c, *lr, *tmp;
	ULONG offset;
	USHORT ind;
	

    irpSp = IoGetCurrentIrpStackLocation(Irp);

    switch (irpSp->Parameters.DeviceIoControl.IoControlCode) {

	case IOCTL_SDLA_DUMMY:
		Irp->IoStatus.Status = STATUS_SUCCESS;
		Irp->IoStatus.Information = sizeof(ULONG);
		tmp = (PULONG) Irp->AssociatedIrp.SystemBuffer;
		break;

	case IOCTL_SDLA_GET_CARDS:
#ifdef DEBG
		DbgPrint("Entering IOCTL_SDLA_GET_CARDS ...\n");
#endif
		c = 0;
		for (cardnumber=0; cardnumber < MAX_CARDS; cardnumber++) {
			if (cardinstances[cardnumber] > 0) c |= (1L << cardnumber);
		}
		Irp->IoStatus.Status = STATUS_SUCCESS;
		Irp->IoStatus.Information = sizeof(ULONG);
		lr = (PULONG) Irp->AssociatedIrp.SystemBuffer;
		*lr = c | ((ULONG)available_cards << 16);
#ifdef DEBG
		DbgPrint("Returning 0x%08X from IOCTL_SDLA_GET_CARDS.\n", *lr);
		DbgPrint("					 available_cards == 0x%X.\n", available_cards);
		sprintf(globalLogString, "Returning 0x%08X from IOCTL_SDLA_GET_CARDS.\0", *lr);
		StartUpLog();
#endif
		break;

	case IOCTL_SDLA_REGISTER:
#ifdef DEBG
		DbgPrint("Entering IOCTL_SDLA_REGISTER ...\n");
#endif
		Irp->IoStatus.Status = STATUS_SUCCESS;
		Irp->IoStatus.Information = sizeof(CARD_TRANSACTION);
		trans = (CARD_TRANSACTION *) Irp->AssociatedIrp.SystemBuffer;
		// get card number and make sure its valid
		if ((cardnumber = trans->cardnumber) >= MAX_CARDS)
		{ 
			trans->rcode = SDLAR_NOT_REGISTERED;
			break; 
		}
		if (blockregistration[cardnumber])
		{
			trans->rcode = SDLAR_CARD_BLOCKED;
			break;
		}
		if (cardinstances[cardnumber] == 0)
		{
			if (!MapMemory(cardnumber))
			{
				trans->rcode = SDLAR_NOT_REGISTERED;
				break;
			}
			cardinstances[cardnumber]++;
			trans->rcode = SDLAR_ALLOCATED;
#ifdef DEBG
   			DbgPrint("Returning %d from IOCTL_SDLA_REGISTER.\n", trans->rcode );
			sprintf(globalLogString, "Returning %d from IOCTL_SDLA_REGISTER.\0", trans->rcode);
			StartUpLog();
#endif	
			break;
		}
		cardinstances[cardnumber]++;
		trans->rcode = SDLAR_REGISTERED;
#ifdef DEBG
    		DbgPrint("SDLAR_REGISTERED:Returning %d from IOCTL_SDLA_REGISTER.\n", trans->rcode );
#endif	
		break;

	case IOCTL_SDLA_DEREGISTER:
#ifdef DEBG
		DbgPrint("Entering IOCTL_SDLA_DEREGISTER ...\n");
#endif
		Irp->IoStatus.Status = STATUS_SUCCESS;
		Irp->IoStatus.Information = sizeof(CARD_TRANSACTION);
		trans = (CARD_TRANSACTION *) Irp->AssociatedIrp.SystemBuffer;
		// get card number and make sure its valid
		if ((cardnumber=trans->cardnumber) >= MAX_CARDS) { 
			trans->rcode = SDLAR_NOT_REGISTERED; break; 
		}
		if (cardinstances[cardnumber] > 0) {
			cardinstances[cardnumber]--;
			if (cardinstances[cardnumber] == 0) {
				UnmapMemory(cardnumber);
				trans->rcode = SDLAR_RELEASED;
#ifdef DEBG
    			DbgPrint("Returning %d from IOCTL_SDLA_DEREGISTER.\n", trans->rcode );
#endif	
				break;
			}
			trans->rcode = SDLAR_DEREGISTERED;
			break;
#ifdef DEBG
    		DbgPrint("Returning %d from IOCTL_SDLA_DEREGISTER.\n", trans->rcode );
#endif	
		}
#ifdef DEBG
   		DbgPrint("Returning %d from IOCTL_SDLA_DEREGISTER.\n", trans->rcode );
		sprintf(globalLogString, "Returning %d from IOCTL_SDLA_DEREGISTER.\0", trans->rcode);
		StartUpLog();
#endif	
		trans->rcode = SDLAR_NOT_REGISTERED;
		break;

	case IOCTL_SDLA_PREPARE_TO_LOAD:
#ifdef DEBG
		DbgPrint("IOCTL_SDLA_PREPARE_TO_LOAD start\n");
#endif	
		Irp->IoStatus.Status = STATUS_SUCCESS;
		Irp->IoStatus.Information = sizeof(CARD_TRANSACTION);
		trans = (CARD_TRANSACTION *) Irp->AssociatedIrp.SystemBuffer;
		// get card number and make sure its valid
		if ((cardnumber=trans->cardnumber) >= MAX_CARDS) { 
			trans->rcode = SDLAR_NOT_REGISTERED; break; 
		}
#ifdef DEBG
		DbgPrint("IOCTL_SDLA_PREPARE_TO_LOAD:hw_type[%d] == %d\n", cardnumber, hw_type[cardnumber]);

		sprintf(globalLogString, "IOCTL_SDLA_PREPARE_TO_LOAD:hw_type[%d] == %d\0", cardnumber, hw_type[cardnumber]);
		StartUpLog();
		//cardmem[cardnumber]
		sprintf(globalLogString, "cardmem[%d] = 0x%x.\0", cardnumber, cardmem[cardnumber]);
		StartUpLog();
#endif	
		if (blockregistration[cardnumber])
			trans->rcode = SDLAR_CARD_BLOCKED;
		else
			if (cardinstances[cardnumber] == 0)
			{
				if (!MapMemory(cardnumber))
					trans->rcode = SDLAR_ERROR_PREPARING;
				else
				{
					blockregistration[cardnumber] = 1;
					trans->rcode = SDLAR_PREPARED;
					// here important addtion: It seems like
					// when working with 502E/503 board
					// its absolutely nessesary to write 0x00
					// out to base+3 - otherwise it'll fail
					// sometimes -A-
					if (hw_type[cardnumber] != 514)
						WRITE_PORT_UCHAR((PUCHAR)(portmem[cardnumber]+3), 0x00);
				}
			}
			else
				trans->rcode = SDLAR_REGISTERED;
		break;

	case IOCTL_SDLA_LOAD_COMPLETE:
#ifdef DEBG
		DbgPrint("IOCTL_SDLA_LOAD_COMPLETE\n");
#endif	
		Irp->IoStatus.Status = STATUS_SUCCESS;
		Irp->IoStatus.Information = sizeof(CARD_TRANSACTION);
		trans = (CARD_TRANSACTION *) Irp->AssociatedIrp.SystemBuffer;
		// get card number and make sure its valid
		if ((cardnumber=trans->cardnumber) >= MAX_CARDS)
		{ 	trans->rcode = SDLAR_NOT_REGISTERED;
			break; 
		}
		if (blockregistration[cardnumber])
		{
			UnmapMemory(cardnumber);
			blockregistration[cardnumber] = 0;
			trans->rcode = SDLAR_RELEASED;
		}
		else
			trans->rcode = SDLAR_NOT_PREPARED;
		break;

	case IOCTL_SDLA_READ_BYTE:
//#ifdef DEBG
//		DbgPrint("IOCTL_SDLA_READ_BYTE\n");
//#endif	
		OutputBuffer = (XFER_BYTE_CONTROL *) Irp->AssociatedIrp.SystemBuffer;
		cardnumber = OutputBuffer->cardnumber;
		offset = OutputBuffer->offset;
		if(hw_type[cardnumber] != 514) 
			OutputBuffer->value = READ_REGISTER_UCHAR((PUCHAR)(cardmem[cardnumber]+MapSetZ80Window(offset, cardnumber)));
		else
		{
			OutputBuffer->value = READ_REGISTER_UCHAR((PUCHAR)(cardmem[cardnumber]+offset));
//#ifdef DEBG
//			sprintf(globalLogString, "READ_BYTE:value = 0x%x offset = 0x%x cardmem[%d] = 0x%x\0",
//										OutputBuffer->value, offset, cardnumber, cardmem[cardnumber]);
//			StartUpLog();
//#endif
		}
		Irp->IoStatus.Status = STATUS_SUCCESS;
		Irp->IoStatus.Information = sizeof(XFER_BYTE_CONTROL);
		break;

	case IOCTL_SDLA_WRITE_BYTE:
		OutputBuffer = (XFER_BYTE_CONTROL *) Irp->AssociatedIrp.SystemBuffer;
		cardnumber = OutputBuffer->cardnumber;
		offset = OutputBuffer->offset;
//#ifdef DEBG
//		DbgPrint("IOCTL_SDLA_WRITE_BYTE offset = 0x%x\n", offset);
//#endif	
		if(hw_type[cardnumber] != 514) 
			WRITE_REGISTER_UCHAR((PUCHAR)(cardmem[cardnumber]+MapSetZ80Window(offset, cardnumber)), OutputBuffer->value);
		else
		{
			WRITE_REGISTER_UCHAR((PUCHAR)(cardmem[cardnumber]+offset), OutputBuffer->value);
//#ifdef DEBG
//			sprintf(globalLogString, "WRITE_BYTE:value = 0x%x offset = 0x%x cardmem[%d] = 0x%x\0",
//										OutputBuffer->value, offset, cardnumber, cardmem[cardnumber]);
//			StartUpLog();
//#endif

		}
		Irp->IoStatus.Status = STATUS_SUCCESS;
		Irp->IoStatus.Information = sizeof(XFER_BYTE_CONTROL);
		break;

	case IOCTL_START_S514:
		OutputBuffer = (XFER_BYTE_CONTROL *) Irp->AssociatedIrp.SystemBuffer;
		cardnumber = OutputBuffer->cardnumber;
#ifdef DEBG
		DbgPrint("IOCTL_START_S514:cardnumber = %d\n", cardnumber);
#endif	
		if(stop_start_s514_register[cardnumber] != NULL)
			WRITE_REGISTER_UCHAR((PUCHAR)(stop_start_s514_register[cardnumber]), S514_CPU_START);//S514_CPU_START = 0x01	
		Irp->IoStatus.Status = STATUS_SUCCESS;
		Irp->IoStatus.Information = sizeof(XFER_BYTE_CONTROL);
		break;

	case IOCTL_STOP_S514:
		OutputBuffer = (XFER_BYTE_CONTROL *) Irp->AssociatedIrp.SystemBuffer;
		cardnumber = OutputBuffer->cardnumber;
#ifdef DEBG
		DbgPrint("IOCTL_STOP_S514:cardnumber = %d\n", cardnumber);
#endif	
		if(stop_start_s514_register[cardnumber] != NULL)
			WRITE_REGISTER_UCHAR((PUCHAR)(stop_start_s514_register[cardnumber]), S514_CPU_HALT);//S514_CPU_HALT = 0x00	
		Irp->IoStatus.Status = STATUS_SUCCESS;
		Irp->IoStatus.Information = sizeof(XFER_BYTE_CONTROL);
		break;

	case IOCTL_SDLA_READ_BLOCK:
		block = (SDLA_BLOCKOP *) Irp->AssociatedIrp.SystemBuffer;
		cardnumber = block->card_number;
		offset = block->offset;
#ifdef DEBG
		DbgPrint("IOCTL_SDLA_READ_BLOCK offset = 0x%x\n", offset);
		DbgPrint("IOCTL_SDLA_READ_BLOCK num_bytes = 0x%x\n", block->num_bytes);
#endif	
		if(hw_type[cardnumber] != 514) 
			READ_REGISTER_BUFFER_UCHAR((PUCHAR)(cardmem[cardnumber]+MapSetZ80Window(offset, cardnumber)), block->buffer, (ULONG)block->num_bytes);
		else
			READ_REGISTER_BUFFER_UCHAR((PUCHAR)(cardmem[cardnumber]+offset), block->buffer, (ULONG)block->num_bytes);

		Irp->IoStatus.Status = STATUS_SUCCESS;
		Irp->IoStatus.Information = sizeof(SDLA_BLOCKOP);
		break;

	case IOCTL_SDLA_WRITE_BLOCK:
		block = (SDLA_BLOCKOP *) Irp->AssociatedIrp.SystemBuffer;
		cardnumber = block->card_number;
		offset = block->offset;
#ifdef DEBG
		DbgPrint("IOCTL_SDLA_WRITE_BLOCK offset = 0x%x\n", offset);
#endif	
		if(hw_type[cardnumber] != 514)
			WRITE_REGISTER_BUFFER_UCHAR((PUCHAR)(cardmem[cardnumber]+MapSetZ80Window(offset, cardnumber)), block->buffer, (ULONG)block->num_bytes);
		else
			WRITE_REGISTER_BUFFER_UCHAR((PUCHAR)(cardmem[cardnumber]+offset), block->buffer, (ULONG)block->num_bytes);

		Irp->IoStatus.Status = STATUS_SUCCESS;
		Irp->IoStatus.Information = sizeof(SDLA_BLOCKOP);
		break;

	case IOCTL_SDLA_ZERO_WINDOW:
#ifdef DEBG
		DbgPrint("IOCTL_SDLA_ZERO_WINDOW\n");
#endif	
		Irp->IoStatus.Status = STATUS_SUCCESS;
		Irp->IoStatus.Information = sizeof(CARD_TRANSACTION);
		trans = (CARD_TRANSACTION *) Irp->AssociatedIrp.SystemBuffer;
		cardnumber = trans->cardnumber;
//#ifdef DEBG
//		DbgPrint("cardnumber == %d\n", cardnumber);
//		DbgPrint("cardmem[cardnumber]== 0x%x\n", cardmem[cardnumber]);
//#endif
//		for (i=0; i <= 0x0e; i+=2)
//		{
//			MapSetZ80Window(i << 12);
//			APIZEROWINDOW(card);
//		}
		if (hw_type[cardnumber] == 508)
		{
			for (ind=0; ind <= 0x0e; ind+=2) 
			{
				MapSetZ80Window((ind << 12), cardnumber);
				for (offset=0; offset < 0x2000; offset+=4)
					WRITE_REGISTER_ULONG((PULONG)(cardmem[cardnumber]+offset), 0UL);
			}
		}
		else
		{
			if(stop_start_s514_register[cardnumber] != NULL)
			{
				for (offset=0; offset < MAX_SIZEOF_S514_MEMORY; offset++)
					WRITE_REGISTER_UCHAR((cardmem[cardnumber]+offset), 0UL);
			}
		}
		
		break;

	case IOCTL_SDLA_READ_PORT_BYTE:
#ifdef DEBG
		DbgPrint("IOCTL_SDLA_READ_PORT_BYTE\n");
#endif	
		OutputBuffer = (XFER_BYTE_CONTROL *) Irp->AssociatedIrp.SystemBuffer;
		cardnumber = OutputBuffer->cardnumber;
		if (hw_type[cardnumber] == 508)
		{
			offset = OutputBuffer->offset;
			OutputBuffer->value = READ_PORT_UCHAR((PUCHAR)(portmem[cardnumber]+offset));
		}
		Irp->IoStatus.Status = STATUS_SUCCESS;
		Irp->IoStatus.Information = sizeof(XFER_BYTE_CONTROL);

		break;

	case IOCTL_SDLA_WRITE_PORT_BYTE:
#ifdef DEBG
		DbgPrint("IOCTL_SDLA_WRITE_PORT_BYTE\n");
#endif	
		OutputBuffer = (XFER_BYTE_CONTROL *) Irp->AssociatedIrp.SystemBuffer;
		cardnumber = OutputBuffer->cardnumber;
		if (hw_type[cardnumber] == 508)
		{
			offset = OutputBuffer->offset;
			WRITE_PORT_UCHAR((PUCHAR)(portmem[cardnumber]+offset), OutputBuffer->value);
		}
		Irp->IoStatus.Status = STATUS_SUCCESS;
		Irp->IoStatus.Information = sizeof(XFER_BYTE_CONTROL);
		break;
    }

    status = Irp->IoStatus.Status;
    IoCompleteRequest(Irp, 0);
    return(status);
}


ULONG MapSetZ80Window(ULONG logical_offset, UCHAR cardnumber)
{
	ULONG physoff;
	UCHAR winnum;
	static UCHAR last_winnum = 0xff;

	physoff = logical_offset % reserve_window_size[cardnumber];

	winnum = (UCHAR)(logical_offset / reserve_window_size[cardnumber]);

	if (last_winnum != winnum) {
		// only set the window if changing it from last time
		WRITE_PORT_UCHAR((PUCHAR)(portmem[cardnumber]+OFF_Z80_WIN_SELECT), winnum);
		last_winnum = winnum;
	}
	return(physoff);
}


UCHAR MapMemory(UCHAR cardnumber)
{
	PHYSICAL_ADDRESS addresslow;
	ULONG bytes;
	UNICODE_STRING path1;
	WCHAR tw[80];
	ULONG base;

	#ifdef DEBG
	DbgPrint("MapMemory()\n");
	#endif

	if (cardnumber >= MAX_CARDS) return(FALSE);

	if (hw_type[cardnumber] == 508)
	{
		FormRegistryPath(&path1, tw, cardnumber);
		if (RtlQueryRegistryValues(RTL_REGISTRY_SERVICES, path1.Buffer, &rtable[0], NULL, NULL)
			!= STATUS_SUCCESS) return(0);

		base = (temp_window_segment << 4) + temp_window_offset;

		addresslow = RtlConvertUlongToLargeInteger((ULONG)base);
		bytes = (ULONG)temp_window_size;
		cardmem_bytes[cardnumber] = bytes;

		cardmem[cardnumber] = MmMapIoSpace(addresslow, bytes, FALSE);
		if (cardmem[cardnumber] == NULL) {
			return(0);
		}

		addresslow = RtlConvertUlongToLargeInteger((ULONG)temp_io_port);
		bytes = 0x04UL;

		portmem[cardnumber] = MmMapIoSpace(addresslow, bytes, FALSE);
		if (portmem[cardnumber] == NULL) {
			MmUnmapIoSpace(cardmem[cardnumber], (ULONG)temp_window_size);
			return(0);
		}

		return(1);
	}
	else
	{
		if(!InitializeCard514(cardnumber))
			return(0);
		else
			return(1);
	}
}

void UnmapMemory(UCHAR cardnumber)
{
	if (hw_type[cardnumber] == 508)
	{
		#ifdef DEBG
		DbgPrint("508 Unmapping Memory\n");
		#endif	
		MmUnmapIoSpace(cardmem[cardnumber], cardmem_bytes[cardnumber]);
		MmUnmapIoSpace(portmem[cardnumber], 0x04UL);
	}
}

void FormRegistryPath(PUNICODE_STRING p, PWCH buffer, UCHAR cardnumber)
{
	buffer[0] = 0;
	RtlInitUnicodeString(p, buffer);
	p->MaximumLength = 79;
	p->Length = 0;

	RtlAppendUnicodeToString(p, L"SDLA\\Parameters\\");
	RtlAppendUnicodeToString(p, fool_proof[cardnumber]);
}

BOOLEAN ReserveResources(IN PDRIVER_OBJECT DriverObject)
{
	PCM_RESOURCE_LIST Resources;
	PCM_PARTIAL_RESOURCE_LIST prl;
	PCM_PARTIAL_RESOURCE_DESCRIPTOR prd;
	ULONG al = 0, ali = 0, i = 0, j = 0, nump = 0, numm = 0;
	BOOLEAN ConflictDetected;
    NTSTATUS status;

    PPCI_COMMON_CONFIG  configInfo = NULL;
	ULONG	LocalBaseMemoryAddress;		//memory on 514
	ULONG	LocalBaseRegisterAddress;	//to enable memory on 514

	// scan through resources and pick out the unique ones to reserve, and how many in total
	nump = numm = 0;
	for (i=0; i < MAX_CARDS; i++) 
	{
		if (available_cards & (0x0001 << i))
		{
			if(hw_type[i] == 508 && AutoConfigS514 == 0)//if AutoConfigS514 == 1 don't do anything for ISA cards
			{	// 508 card is configured
				nump++;
				numm++;
			}

			// scan previous ones to see if there's a duplicate
			if (i > 0)
			{
				for (j=0; j < i; j++)
				{
					if (hw_type[i] == 508)
					{
						if (reserve_port[j] == reserve_port[i]) {
							nump--;
							reserve_port[i] = 0;	// don't use this one
						}
						if (reserve_mem[j] == reserve_mem[i]) {
							numm--;
							reserve_mem[i] = 0;		// don't use this one
						}
					}

					if (hw_type[i] == 514)
					{
						if ((reserve_bus[j] == reserve_bus[i]) &&
							(reserve_slot[j] == reserve_slot[i]) &&
							(reserve_CPU[j] == reserve_CPU[i]))
						{
							//reserve_slot[i] = 1000;
							dont_use[i] = TRUE; // don't use this one
						}
					}
				}
			}
		}
	}


	if (nump)//resources for ISA cards
	{
		#ifdef DEBG
		DbgPrint("ISA:ReserveResources() (number of ports) nump = %d\n", nump);
		#endif
		//
		// Set up the resource requirements list and use the same one for every
		// resource registration we do.
		// The idea is to allocate size of IO_RESOURCE_REQUIREMENTS_LIST which includes one
		// IO_RESOURCE_LIST structure which includes one IO_RESOURCE_DESCRIPTOR; we need to
		// add (nump+numm-1) more descriptors to the list.
		//
		Resources = (PCM_RESOURCE_LIST)ExAllocatePool(
				PagedPool,
				(al=sizeof(CM_RESOURCE_LIST)+(nump+numm-1)*sizeof(CM_PARTIAL_RESOURCE_DESCRIPTOR)) );

		RtlZeroMemory(Resources, al);
	
		Resources->Count = 1;
		Resources->List[0].InterfaceType = Isa;
		Resources->List[0].BusNumber = 0;
		prl = &Resources->List[0].PartialResourceList;
		prl->Count = (nump+numm);
		prd = prl->PartialDescriptors;
	
		// Set up I/O Port resources
		j=0;
		for (i=0; i < nump; i++)
		{
			prd[i].Type = CmResourceTypePort;
			prd[i].ShareDisposition = CmResourceShareDeviceExclusive;
			prd[i].Flags = CM_RESOURCE_PORT_IO;
			prd[i].u.Port.Length = 4;
			while (reserve_port[j] == 0) j++;
	   		prd[i].u.Port.Start = RtlConvertUlongToLargeInteger(reserve_port[j]);
	   		j++;
		}
	
		// Set up Memory resources
		j=0;
		for (i=nump; i < (nump+numm); i++) 
		{
			prd[i].Type = CmResourceTypeMemory;
			prd[i].ShareDisposition = CmResourceShareDeviceExclusive;
			prd[i].Flags = CM_RESOURCE_MEMORY_READ_WRITE;
			prd[i].u.Memory.Length = 0x2000;
			while (reserve_mem[j] == 0) j++;
	   		prd[i].u.Memory.Start = RtlConvertUlongToLargeInteger(reserve_mem[j]);
	   		j++;
		}

		status = IoReportResourceUsage(NULL, DriverObject, Resources, al, DriverObject->DeviceObject,
																	NULL, 0, FALSE, &ConflictDetected);
   		if (status != STATUS_SUCCESS || ConflictDetected)
		{	ExFreePool(Resources);
			#ifdef DEBG
		    DbgPrint("ISA:Failed to assign resources %x or conflict %x. Deleting device and symbolic link...\n",
																								status, ConflictDetected);
			#endif
			sprintf(globalLogString, 
					"ISA:Failed to assign resources %x or conflict %x. Deleting device and symbolic link...\0",
																						status, ConflictDetected);
			StartUpLog();

			return FALSE;
   		}
   	
		#ifdef DEBG
		DbgPrint("ISA:ReserveResources() returning TRUE.\n");
		#endif

		ExFreePool(Resources);
	}//for ISA cards


	{
		#ifdef DEBG
		DbgPrint("514:ReserveResources()\n");
		#endif
		//s514 driver entry
		//
		// Set up the resource requirements list and use the same one for every
		// resource registration we do.
		// The idea is to allocate size of IO_RESOURCE_REQUIREMENTS_LIST which includes one
		// IO_RESOURCE_LIST structure which includes one IO_RESOURCE_DESCRIPTOR
		//
		Resources = (PCM_RESOURCE_LIST)ExAllocatePool(PagedPool, (al=sizeof(CM_RESOURCE_LIST)));

		RtlZeroMemory(Resources, al);
		//
		// Allocate space for a PCI configuration information structure
		//
		configInfo = ExAllocatePoolWithTag(PagedPool, sizeof(PCI_COMMON_CONFIG), 'SANG');

		if (!configInfo)
		{
			#ifdef DEBG
			DbgPrint("514:could not allocate ConfigInfo struct!\n");
			#endif

			sprintf(globalLogString, "could not allocate ConfigInfo PCI struct!!.\0");
			StartUpLog();
			return FALSE;
		}
	
		for (i=0; i < MAX_CARDS; i++)
		{
			if( (hw_type[i] == 514) && (dont_use[i] != TRUE) )
			{
				Resources->Count = 1;
				Resources->List[0].InterfaceType = PCIBus;
				Resources->List[0].BusNumber = reserve_bus[i];
				prl = &Resources->List[0].PartialResourceList;
				prl->Count = 1;
				prd = prl->PartialDescriptors;
	
				// Set up Memory resource
				prd[0].Type = CmResourceTypeMemory;
				prd[0].ShareDisposition = CmResourceShareDeviceExclusive;
				prd[0].Flags = CM_RESOURCE_MEMORY_READ_WRITE;
				//prd[0].u.Memory.Length = 0x100000;
				prd[0].u.Memory.Length = MAX_SIZEOF_S514_MEMORY;

				HalGetBusData(PCIConfiguration,
								reserve_bus[i],//BusNumber,
								reserve_slot[i],//SlotNumber,
								configInfo,
								sizeof(PCI_COMMON_CONFIG));
			
				if((configInfo->VendorID != V3_VENDOR_ID) || (configInfo->DeviceID != V3_DEVICE_ID) || 
												(configInfo->u.type0.SubVendorID != SANGOMA_SUBSYS_VENDOR))
				{
					#ifdef DEBG
					DbgPrint("could not find S514 at Bus %d. SlotNumber %d!!!.\n", reserve_bus[i], reserve_slot[i]);
					#endif

					sprintf(globalLogString, "could not find S514 at Bus %d. SlotNumber %d!!!.\0", 
																	reserve_bus[i], reserve_slot[i]);
					StartUpLog();
					return FALSE;
				}

				#ifdef DEBG
				DbgPrint("Found S514 at Bus %d. SlotNumber %d.\n", reserve_bus[i], reserve_slot[i]);
				DbgPrint("Vendor id = 0x%0x, Device id = 0x%0x, SubVendorID = 0x%X\n",
					(int)configInfo->VendorID, (int)configInfo->DeviceID, configInfo->u.type0.SubVendorID);
				
				SANGPrintConfig(configInfo);
				#endif

				//memory resource[0] - hardware register on s514
				LocalBaseRegisterAddress = configInfo->u.type0.BaseAddresses[0];

				if(reserve_CPU[i] == 0)														
				{																	
					LocalBaseMemoryAddress = configInfo->u.type0.BaseAddresses[1];  //memory resource[1] - for CPU A
					#ifdef DEBG
					DbgPrint("CPU==0:LocalBaseMemoryAddress= 0x%08X\n", LocalBaseMemoryAddress);
					DbgPrint("CPU==0:LocalBaseRegisterAddress= 0x%08X\n", LocalBaseRegisterAddress);
					#endif
				}
				else
				{
					LocalBaseMemoryAddress = configInfo->u.type0.BaseAddresses[2];	//memory resource[2] - for CPU B
					#ifdef DEBG
					DbgPrint("CPU==1:LocalBaseMemoryAddress= 0x%08X\n", LocalBaseMemoryAddress);
					DbgPrint("CPU==1:LocalBaseRegisterAddress= 0x%08X\n", LocalBaseRegisterAddress);
					#endif
				}
				//  We are not using interrupts.
				//	extension->interruptLevel = configInfo->u.type0.InterruptLine;
			    //  extension->interruptVector = configInfo->u.type0.InterruptLine;

				prd[0].u.Memory.Start = RtlConvertUlongToLargeInteger(LocalBaseMemoryAddress);

				#ifdef DEBG
				DbgPrint("514:Calling IoReportResourceUsage() with mem=%x...\n", prd[0].u.Memory.Start);
			    for (ali=0; ali < al; ali++)
			    	DbgPrint("%02X ", *((UCHAR *)Resources+ali));
			    DbgPrint("\n");
				#endif

				status = IoReportResourceUsage(NULL, DriverObject, NULL, 0, DriverObject->DeviceObject,
																	Resources, al, FALSE, &ConflictDetected);
			    if (status != STATUS_SUCCESS || ConflictDetected)
				{
					#ifdef DEBG
				    DbgPrint("514:Failed to assign resources %x or conflict %x. Deleting device and symbolic link...\n",
																								status, ConflictDetected);
					#endif
					sprintf(globalLogString, 
						"S514:Failed to assign resources %x or conflict %x. Deleting device and symbolic link...\0",
																						status, ConflictDetected);
					StartUpLog();

					ExFreePool(Resources);
					return FALSE;
				}
				else
				{	//PUCHAR cardmem[MAX_CARDS];
					//PUCHAR portmem[MAX_CARDS];
					cardmem[i] = (PUCHAR)LocalBaseMemoryAddress;
					portmem[i] = (PUCHAR)LocalBaseRegisterAddress;
					#ifdef DEBG
					DbgPrint("514:Successfully allocated resources.\n");
					// For debugging purposes, display the resources that were allocated for our use.
					SANGPrintResourceList(Resources);
					#endif
				}
			}//if( (hw_type[i] == 514) && (dont_use[i] != TRUE) )
		}//for (i=0; i < MAX_CARDS; i++)
		ExFreePool(Resources);
	}//for PCI cards
	
	return TRUE;
}

void FreeResources(IN PDRIVER_OBJECT DriverObject)
{
	CM_RESOURCE_LIST ar;
	BOOLEAN cd;

	// Need to free the resources
    ar.Count = 0;
   	IoReportResourceUsage(NULL, DriverObject, &ar, sizeof(CM_RESOURCE_LIST), 
										DriverObject->DeviceObject, NULL, 0, FALSE, &cd);
}


BOOLEAN InitializeCard514(UCHAR cardnumber)
{
	ULONG MemLengthInBytes, RegisterLengthInBytes;
    NTSTATUS code;

    ULONG addressSpace;
    PHYSICAL_ADDRESS MemAddress, RegAddress, CtrlRegAddress;

	USHORT c1, i;
	USHORT DelayCount;

	PHYSICAL_ADDRESS memoryStart;		// 8-byte physical addresses.
    ULONG memoryLength;

	PHYSICAL_ADDRESS RegMemoryStart;    // 8-byte physical addresses.
    ULONG RegMemoryLength;

	PHYSICAL_ADDRESS CtrlRegisterAddress; // 8-byte physical addresses.
	ULONG CtrlRegisterLengInBytes;

	ULONG intsdword, ind, PCI_interrupt_configuration;

	#ifdef DEBG
	DbgPrint("InitializeCard514\n");
	#endif					
    memoryStart.HighPart = 0;
    memoryStart.LowPart  =  (ULONG)cardmem[cardnumber];//LocalBaseMemoryAddress;
	memoryLength =  MAX_SIZEOF_S514_MEMORY;

    RegMemoryStart.HighPart = 0;
    RegMemoryStart.LowPart  =  (ULONG)portmem[cardnumber];//LocalBaseRegisterAddress;
    RegMemoryLength =  0x100;

    CtrlRegisterAddress.HighPart = 0;
    CtrlRegisterAddress.LowPart  = (ULONG)cardmem[cardnumber] + S514_CTRL_REG_BYTE;//LocalBaseMemoryAddress + S514_CTRL_REG_BYTE;
    CtrlRegisterLengInBytes =  0x10;
    //
    // Indciate that the resigsters are in memory space on this card
    //
	addressSpace = 0x0;
    //
    // Get the HAL to translate our bus-relative port base address, to an
    // unambiguous address to be used for device access.  If we're returned
    // FALSE, the call failed.
    //
	//ULONG reserve_bus[MAX_CARDS];
	//ULONG reserve_slot[MAX_CARDS];
	//ULONG reserve_CPU[MAX_CARDS];

	//was configured before
	if (stop_start_s514_register[cardnumber] != NULL) return TRUE;

	//This is for memory access
    if (!HalTranslateBusAddress(PCIBus,
                               reserve_bus[cardnumber],//devExt->BusNumber,
                               memoryStart,
                               &addressSpace,
                               &MemAddress))  {
		#ifdef DEBG
        DbgPrint("HalTranslateBusAddress for mem access failed\n");
		#endif
		sprintf(globalLogString, "S514:HalTranslateBusAddress for mem access failed.\0");
		StartUpLog();

        return FALSE;
    }
	
	//This is for enabling memory
    if (!HalTranslateBusAddress(PCIBus,
                               reserve_bus[cardnumber],//devExt->BusNumber,
                               RegMemoryStart,
                               &addressSpace,
                               &RegAddress))  {
		#ifdef DEBG
        DbgPrint("HalTranslateBusAddress for mem enabling failed\n");
		#endif
		sprintf(globalLogString, "S514:HalTranslateBusAddress for mem enabling failed.\0");
		StartUpLog();

        return FALSE;
    }

	//This is for memory control register
    if (!HalTranslateBusAddress(PCIBus,
                               reserve_bus[cardnumber],//devExt->BusNumber,
                               CtrlRegisterAddress,
                               &addressSpace,
                               &CtrlRegAddress))  {
		#ifdef DEBG
        DbgPrint("HalTranslateBusAddress for CtrlRegisterAddress failed\n");
		#endif
		sprintf(globalLogString, "S514:HalTranslateBusAddress for CtrlRegisterAddress failed.\0");
		StartUpLog();

        return FALSE;
    }

#ifdef DEBG
	DbgPrint("MemAddress.LowPart = 0x%X MemAddress.HighPart = 0x%X\n", MemAddress.LowPart, MemAddress.HighPart);	
	sprintf(globalLogString, "MemAddress.LowPart = 0x%X MemAddress.HighPart = 0x%X\n\0", 
																		MemAddress.LowPart, MemAddress.HighPart);	
	StartUpLog();

	DbgPrint("RegAddress.LowPart = 0x%X RegAddress.HighPart = 0x%X\n", RegAddress.LowPart, RegAddress.HighPart);	
	sprintf(globalLogString, "RegAddress.LowPart = 0x%X RegAddress.HighPart = 0x%X\0",
																		RegAddress.LowPart, RegAddress.HighPart);	
	StartUpLog();

	DbgPrint("CtrlRegAddress.LowPart = 0x%X CtrlRegAddress.HighPart = 0x%X\0",
													CtrlRegAddress.LowPart, CtrlRegAddress.HighPart);	
	sprintf(globalLogString, "CtrlRegAddress.LowPart = 0x%X CtrlRegAddress.HighPart = 0x%X\n", 
													CtrlRegAddress.LowPart, CtrlRegAddress.HighPart);	
	StartUpLog();
#endif

	//
    // This card's I/O space registers actually appear in memory space on
    // this processor, we need to map them with some kernel virtual addresses
    // so we can access them. 
	//
    //
    // Maps a potentially 64 bit physical address to a 32 bit virtual address.
    //
	//devExt->BaseMemoryAddress = (ULONG)MmMapIoSpace(MemAddress, MAX_SIZEOF_S514_MEMORY, FALSE);
	cardmem[cardnumber] = MmMapIoSpace(MemAddress, MAX_SIZEOF_S514_MEMORY, FALSE);
	if (cardmem[cardnumber] == NULL)
	{	sprintf(globalLogString, "MmMapIoSpace() for cardmem[cardnumber] (BaseMemoryAddress) failed.\0");	
		StartUpLog();
		return FALSE;
	}
	#ifdef DEBG
	sprintf(globalLogString, "cardmem[%d] = 0x%x\n\0", cardnumber, cardmem[cardnumber]);	
	StartUpLog();
	#endif

    //devExt->BaseRegisterAddress = (ULONG)MmMapIoSpace(RegAddress, RegMemoryLength, FALSE);
	portmem[cardnumber] = MmMapIoSpace(RegAddress, RegMemoryLength, FALSE);
	if (portmem[cardnumber] == NULL)
	{	sprintf(globalLogString, "MmMapIoSpace() for portmem[cardnumber] (BaseRegisterAddress) failed.\0");	
		StartUpLog();
		return FALSE;
	}
	#ifdef DEBG
	sprintf(globalLogString, "portmem[%d] = 0x%x\n\0", cardnumber, portmem[cardnumber]);	
	StartUpLog();
	#endif

	//devExt->Memory_S514_CTRL_REG_BYTE = (ULONG)MmMapIoSpace(CtrlRegAddress, CtrlRegisterLengInBytes, FALSE);
	stop_start_s514_register[cardnumber] = MmMapIoSpace(CtrlRegAddress, CtrlRegisterLengInBytes, FALSE);
	if (stop_start_s514_register[cardnumber] == NULL)
	{	sprintf(globalLogString, "MmMapIoSpace() for stop_start_s514_register[cardnumber] (Memory_S514_CTRL_REG_BYTE) failed.\0");	
		StartUpLog();
		return FALSE;
	}
	#ifdef DEBG
	sprintf(globalLogString, "stop_start_s514_register[%d] = 0x%x\n\0", cardnumber, stop_start_s514_register[cardnumber]);	
	StartUpLog();

	DbgPrint("Memory_S514_CTRL_REG_BYTE =0x%x\n", stop_start_s514_register[cardnumber]);
	#endif

	#ifdef DEBG
	//0x00 0x2C 0x4C
	DbgPrint("At portmem[cardnumber] : 0x%08X\n", READ_REGISTER_ULONG((PULONG)(portmem[cardnumber])) );
	sprintf(globalLogString, "At portmem[cardnumber] : 0x%08X\n", READ_REGISTER_ULONG((PULONG)(portmem[cardnumber])) );
	StartUpLog();

	DbgPrint("At (portmem[cardnumber] + 0x2C ) : 0x%08X\n", READ_REGISTER_ULONG((PULONG)(portmem[cardnumber] + 0x2C)) );
	sprintf(globalLogString, "At (portmem[cardnumber] + 0x2C ) : 0x%08X\n", READ_REGISTER_ULONG((PULONG)(portmem[cardnumber] + 0x2C)) );
	StartUpLog();

	DbgPrint("At (portmem[cardnumber] + 0x4C ) : 0x%08X\n", READ_REGISTER_ULONG((PULONG)(portmem[cardnumber] + 0x4C)) );
	sprintf(globalLogString, "At (portmem[cardnumber] + 0x4C ) : 0x%08X\n", READ_REGISTER_ULONG((PULONG)(portmem[cardnumber] + 0x4C)) );
	StartUpLog();

	#endif

	if(reserve_CPU[cardnumber] == 0)
	{
		ULONG reg_val = 0x00;
		//
		//Let's enable memory on CPU A. Move to offset PCI_MAP0_DWORD(0x40) from BaseRegisterAddress.
		//
		#ifdef DEBG
		DbgPrint("CPU == 0:s514 memory starts at 0x%x\n", cardmem[cardnumber]);
		DbgPrint("s514 Register at 0x%x\n\n", portmem[cardnumber]);
		#endif

		reg_val = READ_REGISTER_ULONG((PULONG)(portmem[cardnumber] + PCI_MAP0_DWORD));

		#ifdef DEBG
		DbgPrint("Enable Mem register(HAL) Before OR: 0x%08X\n", reg_val);
		sprintf(globalLogString, "Enable Mem register(HAL): 0x%08X\n", reg_val);
		StartUpLog();
		#endif

		reg_val |= PCI_MEMORY_ENABLE;
		WRITE_REGISTER_ULONG((PULONG)(portmem[cardnumber] + PCI_MAP0_DWORD), reg_val);

//		PCI_interrupt_configuration = READ_REGISTER_ULONG((PULONG)(devExt->BaseRegisterAddress + 0x4C));
//		PCI_interrupt_configuration |= 0x00000008;
//		WRITE_REGISTER_ULONG((PULONG)(devExt->BaseRegisterAddress + 0x4C), PCI_interrupt_configuration);

		#ifdef DEBG
		reg_val = READ_REGISTER_ULONG((PULONG)(portmem[cardnumber] + PCI_MAP0_DWORD));
		DbgPrint("Enable Mem register(HAL) after OR: 0x%08X\n", reg_val);
		sprintf(globalLogString, "Enable Mem register(HAL) after OR: 0x%08X\n", reg_val);
		StartUpLog();
		DbgPrint("PCI_interrupt_configuration : 0x%08X\n", *(PULONG)(portmem[cardnumber] + PCI_INT_CONFIG));//0x4C
		#endif
	}
	else
	{
		ULONG reg_val = 0x00;
		//
		//Let's enable memory on CPU B. Move to offset PCI_MAP1_DWORD(0x44) from BaseRegisterAddress.
		//
		#ifdef DEBG
		DbgPrint("CPU == 1:s514 memory starts at 0x%x\n", cardmem[cardnumber]);
		DbgPrint("s514 Register at 0x%x\n\n", portmem[cardnumber]);
		#endif

		reg_val = READ_REGISTER_ULONG((PULONG)(portmem[cardnumber] + PCI_MAP1_DWORD));

		#ifdef DEBG
		DbgPrint("Enable Mem register(HAL) Before OR: 0x%08X\n", reg_val);
		sprintf(globalLogString, "Enable Mem register(HAL): 0x%08X\n", reg_val);
		StartUpLog();
		#endif

		reg_val |= PCI_MEMORY_ENABLE;
		WRITE_REGISTER_ULONG((PULONG)(portmem[cardnumber] + PCI_MAP1_DWORD), reg_val);

//		PCI_interrupt_configuration = READ_REGISTER_ULONG((PULONG)(devExt->BaseRegisterAddress + 0x4C));
//		PCI_interrupt_configuration |= 0x00000008;
//		WRITE_REGISTER_ULONG((PULONG)(devExt->BaseRegisterAddress + 0x4C), PCI_interrupt_configuration);

		#ifdef DEBG
		reg_val = READ_REGISTER_ULONG((PULONG)(portmem[cardnumber] + PCI_MAP1_DWORD));
		DbgPrint("Enable Mem register(HAL) after OR: 0x%08X\n", reg_val);
		sprintf(globalLogString, "Enable Mem register(HAL) after OR: 0x%08X\n", reg_val);
		StartUpLog();
		DbgPrint("PCI_interrupt_configuration : 0x%08X\n", *(PULONG)(portmem[cardnumber] + PCI_INT_CONFIG));//0x4C
		#endif
	}

	if(!DoMemoryTest(cardnumber))
	{	sprintf(globalLogString, "Memory Test failed.\n");
		StartUpLog();
		return FALSE;
	}

	//scan for logical cards using the same phisycal card
	for(i=0; i < MAX_CARDS; i++)
	{
		if ((reserve_bus[cardnumber] == reserve_bus[i]) &&
			(reserve_slot[cardnumber] == reserve_slot[i]) &&
			(reserve_CPU[cardnumber] == reserve_CPU[i]) &&
			(dont_use[cardnumber] == TRUE))
		{
			cardmem[cardnumber] = cardmem[i];
			portmem[cardnumber] = portmem[i];
			stop_start_s514_register[cardnumber] = stop_start_s514_register[i];
			//reserve_slot[i] = 1000; // don't use this one
		}
	}

    return TRUE;
}

BOOLEAN DoMemoryTest(UCHAR cardnumber)
{
	UCHAR indx, test_byte = 0x00;

#ifdef DEBG
	DbgPrint("Doing Mem Test...cardnumber = %d\n", cardnumber);
	DbgPrint("Mapped Memory : cardmem[%d] = 0x%x\n", cardnumber, cardmem[cardnumber]);
#endif

	for(indx = 0x00; indx < 0xFF; indx++)
	{
		WRITE_REGISTER_UCHAR( (PUCHAR)(cardmem[cardnumber]+indx), indx);
#ifdef DEBG
		DbgPrint("WRITE_BYTE:indx = 0x%x \n", indx);
		sprintf(globalLogString, "WRITE_BYTE:indx = 0x%x \0", indx);
		StartUpLog();
#endif
		test_byte = READ_REGISTER_UCHAR( (PUCHAR)(cardmem[cardnumber]+indx) );

#ifdef DEBG
		DbgPrint("READ_BYTE:test_byte = 0x%x \n", test_byte);
		sprintf(globalLogString, "READ_BYTE:test_byte = 0x%x \0", test_byte);
		StartUpLog();
#endif

		if( indx != test_byte )
		{	DbgPrint("MemTest Failed!!:indx = 0x%x test_byte = 0x%x \n", indx, test_byte);
			sprintf(globalLogString, "MemTest Failed!!:indx = 0x%x test_byte = 0x%x \0", indx, test_byte);
			StartUpLog();
			return FALSE;
		}
	}
	return TRUE;
}

//
// Some static string tables we use as part of debugging
//
static PSTR CmResourceTypeStrings[] = 
{
    "CmResourceTypeNull",
    "CmResourceTypePort",
    "CmResourceTypeInterrupt",
    "CmResourceTypeMemory",
    "CmResourceTypeDma",
    "CmResourceTypeDeviceSpecific",
    "CmResourceTypeMaximum"
};

static PSTR CmShareDispositionStrings[] = 
{
    "CmResourceShareUndetermined",
    "CmResourceShareDeviceExclusive",
    "CmResourceShareDriverExclusive",
    "CmResourceShareShared"
};


static VOID SANGPrintConfig(PPCI_COMMON_CONFIG  configInfo)
{
    ULONG index;

    DbgPrint("Displaying PCI Configuration Information\n");
    DbgPrint("\tRevisionID is 0x%x\n", (int)configInfo->RevisionID);
    DbgPrint("\tProgIf is 0x%x\n", (int) configInfo->ProgIf);
    DbgPrint("\tSubClass is 0x%x\n", (int) configInfo->SubClass);
    DbgPrint("\tBaseClass is 0x%x\n", (int) configInfo->BaseClass);
    DbgPrint("\tCacheLineSize is 0x%x\n", (int) configInfo->CacheLineSize);
    DbgPrint("\tLatencyTimer is 0x%x\n", (int) configInfo->LatencyTimer);
    DbgPrint("\tHeaderType is 0x%x\n", (int) configInfo->HeaderType);
    DbgPrint("\tBIST is 0x%x\n", (int) configInfo->HeaderType);

    for (index = 0; index < PCI_TYPE0_ADDRESSES; index++)  {

        DbgPrint("\tBaseAddresses[%d] is 0x%x\n",index, configInfo->u.type0.BaseAddresses[index]);

    }

    DbgPrint("\tROMBaseAddress is 0x%x\n", configInfo->u.type0.ROMBaseAddress);
    DbgPrint("\tInterruptLine is 0x%x\n", configInfo->u.type0.InterruptLine);
    DbgPrint("\tInterruptPin is 0x%x\n", configInfo->u.type0.InterruptPin);
    DbgPrint("****************************\n");

}  


static VOID SANGPrintResourceList(PCM_RESOURCE_LIST Resources)
{
    ULONG index, index2;
        
    DbgPrint("%d. resource descriptor list(s) returned\n", Resources->Count);

    for (index = 0; index < Resources->Count; index++)  {
        
        DbgPrint("\t[%d] Interface Type 0x%x\n",
                index, Resources->List[index].InterfaceType);
        DbgPrint("\t[%d] BusNumber 0x%x\n",
                index, Resources->List[index].BusNumber);
        DbgPrint("\t[%d] Version 0x%x\n",
                index, Resources->List[index].PartialResourceList.Version);
        DbgPrint("\t[%d] Revision 0x%x\n",
                index, Resources->List[index].PartialResourceList.Revision);
        
        DbgPrint("\t[%d] Partial Resource Descriptors %d.\n",
                index, Resources->List[index].PartialResourceList.Count);
        for (index2 = 0;
            index2 < Resources->List[index].PartialResourceList.Count;
            index2++)  {

            PCM_PARTIAL_RESOURCE_DESCRIPTOR prd; // Lots of typing
            
            prd = &Resources->List[index].PartialResourceList.PartialDescriptors[index2];

            DbgPrint("\t\t[%d] Type 0x%x (%s)\n",
                    index2, prd->Type, CmResourceTypeStrings[prd->Type]);
            DbgPrint("\t\t[%d] Share Disposition 0x%x (%s)\n",
                    index2, prd->ShareDisposition,
                    CmShareDispositionStrings[prd->ShareDisposition]);
            DbgPrint("\t\t[%d] Flags 0x%x\n", index2, prd->Flags);
            DbgPrint("\t\t[%d] Raw 0x%x %x %x\n",
                    index2, prd->u.DeviceSpecificData.DataSize,
                    prd->u.DeviceSpecificData.Reserved1,
                    prd->u.DeviceSpecificData.Reserved2);

            switch (prd->Type) {
                
                case CmResourceTypePort:
                    if (prd->Flags == CM_RESOURCE_PORT_MEMORY)
                        DbgPrint("\t\t[%d] port memory starting at 0x%x length 0x%x\n", 
                                      index2, prd->u.Port.Start.LowPart,
                                      prd->u.Port.Length);
                    if (prd->Flags == CM_RESOURCE_PORT_IO)
                       DbgPrint("\t\t[%d] port i/o starting at 0x%x length 0x%x\n", 
                                     index2, prd->u.Port.Start.LowPart,
                                     prd->u.Port.Length);
                    break;

                case CmResourceTypeInterrupt:
                    if (prd->Flags == CM_RESOURCE_INTERRUPT_LEVEL_SENSITIVE)
                        DbgPrint("\t\t[%d] level interrupt at lvl 0x%x vector 0x%x affinity 0x%x\n", 
                                     index2, prd->u.Interrupt.Level,
                                     prd->u.Interrupt.Vector,
                                     prd->u.Interrupt.Affinity);
                    if (prd->Flags == CM_RESOURCE_INTERRUPT_LATCHED)
                        DbgPrint("\t\t[%d] latched interrupt at lvl 0x%x vector 0x%x affinity 0x%x\n", 
                                     index2, prd->u.Interrupt.Level,
                                     prd->u.Interrupt.Vector,
                                     prd->u.Interrupt.Affinity);
                    break;

                case CmResourceTypeMemory:
                    if (prd->Flags == CM_RESOURCE_MEMORY_READ_WRITE)
                        DbgPrint("\t\t[%d] r/w memory starting at 0x%x length 0x%x\n",
                                 index2, prd->u.Memory.Start.LowPart,
                                 prd->u.Memory.Length);
                    if (prd->Flags & CM_RESOURCE_MEMORY_READ_ONLY)
                        DbgPrint("\t\t[%d] r/o memory starting at 0x%x length 0x%x\n",
                                 index2, prd->u.Memory.Start.LowPart,
                                 prd->u.Memory.Length);
                    if (prd->Flags & CM_RESOURCE_MEMORY_WRITE_ONLY)
                        DbgPrint("\t\t[%d] w/o memory starting at 0x%x length 0x%x\n",
                                 index2, prd->u.Memory.Start.LowPart,
                                 prd->u.Memory.Length);
                    break;

                case CmResourceTypeDma:
                    DbgPrint("\t\t[%d] DMA on channel 0x%x\n",
                        index2, prd->u.Dma.Channel);
                    break;

                case CmResourceTypeDeviceSpecific:
                    DbgPrint("\t\t[%d] Device specific data at 0x%x length 0x%x\n",
                                 index2,
                                 ((ULONG) &prd->u.DeviceSpecificData.Reserved2) + (ULONG)sizeof(ULONG),
                                 prd->u.DeviceSpecificData.DataSize);
                    break;

                default:
                    //
                    // Say what?!!  Unknown resource type.  Something is pretty wierd here.
                    //
                    DbgPrint("Unknown resource type 0x%x\n", prd->Type);
                    break;
            }
            
        }
            DbgPrint("\t[%d] ***** End dump ******\n", index);
    }

}

void StartUpLog(void)
{
	NTSTATUS NT_status;
	UNICODE_STRING FullFileName;
	WCHAR * wchar_string = L"\\SystemRoot\\system32\\drivers\\sdlalog.wri";
	HANDLE  NtFileHandle;

	UCHAR str[200];
	PUCHAR ptr_log_str = &globalLogString[0];
	PUCHAR ptr_str = &str[0];

	OBJECT_ATTRIBUTES ObjectAttributes; 
	IO_STATUS_BLOCK IoStatus;
	TIME_FIELDS TimeFieldes;
	LARGE_INTEGER  CurrentSystemTime, CurrentLocalTime;

	RtlInitUnicodeString(&FullFileName, wchar_string);

    InitializeObjectAttributes ( &ObjectAttributes, 
                                &FullFileName, 
                                OBJ_CASE_INSENSITIVE, 
                                NULL, 
								NULL );

	NT_status = ZwCreateFile(   &NtFileHandle, 
	                            SYNCHRONIZE | FILE_READ_DATA | FILE_APPEND_DATA, 
	                            &ObjectAttributes, 
		                        &IoStatus, 
			                    NULL,                          // alloc size = none 
				                FILE_ATTRIBUTE_NORMAL, 
					            FILE_SHARE_READ | FILE_SHARE_WRITE, 
						        FILE_OPEN_IF, 
							    FILE_SYNCHRONOUS_IO_NONALERT, 
								NULL,  // eabuffer 
								0 );   // ealength 

	if ( !NT_SUCCESS( NT_status ) ) 
	{ 
		#ifdef DEBG
		DbgPrint("ZwCreateFile failed\n");
		#endif	
		return;
	} 
			
	KeQuerySystemTime( &CurrentSystemTime );
 
	ExSystemTimeToLocalTime(&CurrentSystemTime,	&CurrentLocalTime );

	RtlTimeToTimeFields( &CurrentLocalTime, &TimeFieldes );

//	#ifdef DEBG
//	DbgPrint("TimeFieldes.Day =%u\n", TimeFieldes.Day);
//	DbgPrint("TimeFieldes.Hour =%u\n", TimeFieldes.Hour);
//	DbgPrint("TimeFieldes.Minute =%u\n", TimeFieldes.Minute);
//	DbgPrint("TimeFieldes.Second =%u\n", TimeFieldes.Second);
//	DbgPrint("TimeFieldes.Weekday =%u\n", TimeFieldes.Weekday);
//	#endif
	memset( str, '\0', 190);

	sprintf(str,"\n%u/%u/%u %u:%u:%u ", TimeFieldes.Day, TimeFieldes.Month, TimeFieldes.Year,
					TimeFieldes.Hour, TimeFieldes.Minute, TimeFieldes.Second);
	strcat(str, ptr_log_str);
	#ifdef DEBG
	DbgPrint(str);
	#endif

	NT_status = ZwWriteFile(  NtFileHandle,
							  NULL,						
							  NULL,						
							  NtFileHandle,				
							  &IoStatus,
							  (PVOID)ptr_str, 
							  strlen(str),
							  NULL,						
							  NULL						
							);

	ZwClose( NtFileHandle );
}

/////////////////////////////////////////////////////////////////////////////////

VOID FindSangomaCard(void)
{
    NTSTATUS code;
    PPCI_COMMON_CONFIG  configInfo = NULL;
    ULONG busNumber, busNum, Local_slotNum;
    ULONG deviceNumber;
    PCI_SLOT_NUMBER slotNumber, slotNum;
    ULONG length, i;
    BOOLEAN moreBuses;
    ULONG index, index2, index3;

    //
    // Allocate space for a PCI configuration information structure
    //
    configInfo = ExAllocatePoolWithTag(PagedPool, sizeof(PCI_COMMON_CONFIG), 'pRSO');

    if (!configInfo)
	{
		#ifdef DEBG
		DbgPrint("FindSangomaCard:ExAllocatePoolWithTag failed!!!\n");
		#endif 
        return;
    }

	//finding S514 cards on PCI buses connected to the system
	//
	//Keep looking at buses untill there are no more

    adapterFound = FALSE;
    moreBuses = TRUE;

	for (busNumber = 0;moreBuses;busNumber++)
	{
		//
		//Enumerate all the devices on this bus
		//
		for (deviceNumber=0;moreBuses && deviceNumber < PCI_MAX_DEVICES;deviceNumber++)
		{
			//
			//Enumerate all functions on each device
			//
			
			slotNumber.u.bits.Reserved=0;
			slotNumber.u.bits.DeviceNumber = deviceNumber;
			slotNumber.u.bits.FunctionNumber = 0;	//one function so far
				
			//
			//Get the configuration space for the adapter in this slot
			//
			length = HalGetBusData( PCIConfiguration,
									busNumber,
									slotNumber.u.AsULONG,
									configInfo,
									sizeof(PCI_COMMON_CONFIG));
			//
			//A return value of zero indicates no more PCI buses on the system
			//
			if (length == 0)
			{
				moreBuses = FALSE;
				break;
			}
			//If there's nothing in this slot, PCI_INVALID_VENDORID is returned
			//as a vendor ID.

			if (configInfo->VendorID == PCI_INVALID_VENDORID)
			{
				continue;
			}

			//If both the vendor and device ID match then this is the device.
			//
			//Also : only two types of cards can be used by this driver S5141 and S5142
			//if some other Sangoma card is found ignore it.
			if( (configInfo->VendorID == V3_VENDOR_ID) &&
				(configInfo->DeviceID == V3_DEVICE_ID) &&
				(configInfo->u.type0.SubVendorID == SANGOMA_SUBSYS_VENDOR) &&
				(configInfo->u.type0.SubSystemID == S5141_ADPTR_1_CPU_SERIAL ||
				configInfo->u.type0.SubSystemID == S5142_ADPTR_2_CPU_SERIAL))
			{
#ifdef DEBG
	DbgPrint("Found S514 at Bus %d. Device %d SlotNumber %d.\n", busNumber, deviceNumber, slotNumber.u.AsULONG);

    DbgPrint("Vendor id = 0x%0x, Device id = 0x%0x\n", (int)configInfo->VendorID, (int)configInfo->DeviceID);

	DbgPrint("SubVendorID = 0x%X\n", configInfo->u.type0.SubVendorID);

	DbgPrint("SubSystemID = 0x%X\n", configInfo->u.type0.SubSystemID);
#endif
				BusNumber = (USHORT)busNumber;
				SlotNumber = (USHORT)slotNumber.u.AsULONG;
				CPU = 0;

				adapterFound = TRUE;

				//initialize all global arrays
				for (i=0; i < MAX_CARDS; i++)
				{
					cardinstances[i] = 0;
					cardmem[i] = NULL;
					blockregistration[i] = 0;
					stop_start_s514_register[i] = NULL;
					dont_use[i] = TRUE;
				}

				//CPU A
				i = 0;
				dont_use[i] = FALSE;
				available_cards |= (0x0001 << i);
				reserve_bus[i] = BusNumber;
				reserve_slot[i] = SlotNumber;
				reserve_CPU[i] = i;
				hw_type[i] = 514;

				if(configInfo->u.type0.SubSystemID == S5142_ADPTR_2_CPU_SERIAL)
				{	//CPU B
					i = 1;
					dont_use[i] = FALSE;
					available_cards |= (0x0001 << i);
					reserve_bus[i] = BusNumber;
					reserve_slot[i] = SlotNumber;
					reserve_CPU[i] = i;
					hw_type[i] = 514;
				}

#ifdef DEBG
				DbgPrint("End of search\n");
#endif
				SangReturnPool(configInfo);
				return;
			}
		}
	}

    return;
}

VOID SangReturnPool(PPCI_COMMON_CONFIG  ConfigInfo)
{
    if(ConfigInfo)
        ExFreePool(ConfigInfo);
}    
